/**CFile***********************************************************************

  FileName    [form.c]

  Package     [form]

  Synopsis    [Formalas manipulation]

  Description [This file provides the function to create and
  manipulate LTL formulas. ]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.9]

******************************************************************************/

#include <stdio.h>
#include "main.h"
#include "struct.h"
#include "form.h"
#include "formInt.h"
#include "prop.h"
#include "modella.y.tab.h" 


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [It creates a new formula. ]

******************************************************************************/
form* new_form(int type,int prop,form* left,form* right){
  form *t;
  list* tmp=formulas;

  /* look if the formula is already present in formulas */
  while(tmp!=NULL){
    t=(form*)tmp->el;
    if(t->type==type && t->prop==prop && t->left==left && t->right==right)
      return t;
    else
      tmp=tmp->next;
  }

  /* create a new formula */
  t=(form*)mymalloc(sizeof(form));
  t->type=type;
  t->prop=prop;
  t->left=left;
  t->right=right;
  t->id=NULL;

  /* if the id variables are initialized, create the id */
  if(ID_IS_INIT)
    creaid(t);

  formulas=new_list(t,formulas);
  return t;
}

/**Function********************************************************************

  Synopsis    [It creates a new formula with assigned id. ]

  SeeAlso     [Xfactoring]

******************************************************************************/
form* new_form_withid(int type,int prop,form* left,form* right,idtype id){
  form *t;
  list* tmp=formulas;
  while(tmp!=NULL){
    t=(form*)tmp->el;
    if(t->type==type && t->prop==prop && t->left==left && t->right==right)
      return t;
    else
      tmp=tmp->next;
  }
  t=(form*)mymalloc(sizeof(form));
  t->type=type;
  t->prop=prop;
  t->left=left;
  t->right=right;
  t->id=id;
  formulas=new_list(t,formulas);
  return t;
}


/**Function********************************************************************

  Synopsis    [It initializes the form global variables.]

******************************************************************************/
void initialize_form(){
  formulas = NULL;
  true = new_form(TRUE,0,0,0);
  false = new_form(FALSE,0,0,0);  
}

/**Function********************************************************************

  Synopsis    [It initializes the fair condition variables.]

******************************************************************************/
void initialize_faircond(){
  int i=0;
  list* tmp=formulas;
  faircond=(fair**)mymalloc(fair_number*sizeof(fair*));
  while(tmp!=NULL){
    form* f=tmp->el;
    if(f->type==UNTIL){
      faircond[i]=(fair*)mymalloc(sizeof(fair));
      faircond[i]->promise=f;
      faircond[i++]->fulfill=f->right;
    }
    tmp=tmp->next;
  }
  fair_number=i;
}

/**Function********************************************************************

  Synopsis    [It returns true if t1 and t2 are equal.]

******************************************************************************/
int arequal(form* t1,form* t2){
  return (t1->type==t2->type && t1->prop==t2->prop && t1->left==t2->left && t1->right==t2->right);
}




void print_form_ind(form* t,int ni){
  int i;
  switch(t->type){
  case TRUE:
    for(i=0;i<ni;i++)
      printf("   ");
    printf("T\n");
    break;
  case FALSE:
    for(i=0;i<ni;i++)
      printf("   ");
    printf("F\n");
    break;
  case ATOM:
    for(i=0;i<ni;i++)
      printf("   ");
    printf("p%d\n",t->prop);
    break;
  case NOT:
    for(i=0;i<ni;i++)
      printf("   ");
    printf("!p%d\n",t->prop);
    break;
  case NEXT:
    for(i=0;i<ni;i++)
      printf("   ");
    printf("X\n");
    print_form_ind(t->left,ni+1);
    break;
  case AND:
    print_form_ind(t->left,ni+1);
    for(i=0;i<ni;i++)
      printf("   ");
    printf("&\n");
    print_form_ind(t->right,ni+1);
    break;
  case OR:
    print_form_ind(t->left,ni+1);
    for(i=0;i<ni;i++)
      printf("   ");
    printf("|\n");
    print_form_ind(t->right,ni+1);
    break;
  case UNTIL:
    print_form_ind(t->left,ni+1);
    for(i=0;i<ni;i++)
      printf("   ");
    printf("U\n");
    print_form_ind(t->right,ni+1);
    break;
  case RELEASE:
    print_form_ind(t->left,ni+1);
    for(i=0;i<ni;i++)
      printf("   ");
    printf("V\n");
    print_form_ind(t->right,ni+1);
    break;
  }
}
    
void print_form_normal(form* t){
  switch(t->type){
  case TRUE:
    printf("T");
    break;
  case FALSE:
    printf("F");
    break;
  case ATOM:
    printf("p%d",t->prop);
    break;
  case NOT:
    printf("!p%d",t->prop);
    break;
  case NEXT:
    printf("X");
    print_form_normal(t->left);
    break;
  case AND:
    printf("(");
    print_form_normal(t->left);
    printf(" & ");
    print_form_normal(t->right);
    printf(")");
    break;
  case OR:
    printf("(");
    print_form_normal(t->left);
    printf(" | ");
    print_form_normal(t->right);
    printf(")");
    break;
  case UNTIL:
    printf("(");
    print_form_normal(t->left);
    printf(" U ");
    print_form_normal(t->right);
    printf(")");
    break;
  case RELEASE:
    printf("(");
    print_form_normal(t->left);
    printf(" V ");
    print_form_normal(t->right);
    printf(")");
    break;
  }
}

void print_form(form* t){
  print_form_normal(t);
  printf("\n");
}


form* DNF(form* t){
  form *t1,*t2,*t3=NULL;
  form *tmp1,*tmp2,*tmp3;
  if(t->type==AND || t->type==OR){
    t1=DNF(t->left);
    t2=DNF(t->right);
  } else
    return t;
  if(t->type==AND){
    stack* s1=new_stack(); // by mean of s1, we visit t->left in DPF
    tmp1=t1;
    while(tmp1->type==OR){
      s1=push(tmp1,s1);
      tmp1=tmp1->left;
    }
    while(tmp1!=NULL){
      stack* s2=new_stack(); // by mean of s1, we visit t->right in DPF
      tmp2=t2;
      while(tmp2->type==OR){
	s2=push(tmp2,s2);
	tmp2=tmp2->left;
      }
      while(tmp2!=NULL){
	tmp3=new_form(AND,0,tmp1,tmp2);
	if(t3==NULL)
	  t3=tmp3;
	else
	  t3=new_form(OR,0,t3,tmp3);
	s2=pop((void**)&tmp3,s2);
	while(tmp3!=NULL && tmp3->right==tmp2){
	  tmp2=tmp3;
	  s2=pop((void**)&tmp3,s2);
	}
	if(tmp3!=NULL){
	  s2=push(tmp3,s2);
	  tmp2=tmp3->right;
	  while(tmp2->type==OR){
	    s2=push(tmp2,s2);
	    tmp2=tmp2->left;
	  }
	} else 
	  tmp2=tmp3;
      }
      s1=pop((void**)&tmp3,s1);
      while(tmp3!=NULL && tmp3->right==tmp1){
	tmp1=tmp3;
	s1=pop((void**)&tmp3,s1);
      }
      if(tmp3!=NULL){
	s1=push(tmp3,s1);
	tmp1=tmp3->right;
	while(tmp1->type==OR){
	  s1=push(tmp1,s1);
	  tmp1=tmp1->left;
	}
      } else 
	tmp1=tmp3;
    }
  return t3;
  } else return new_form(OR,0,t1,t2);
}
  
form* expand(form* t){
  switch(t->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
  case NEXT:
    return t;
  case AND:
    return new_form(AND,0,expand(t->left),expand(t->right));
  case OR:
    return new_form(OR,0,expand(t->left),expand(t->right));
  case UNTIL:
    return new_form(OR,0,expand(t->right),new_form(AND,0,expand(t->left),new_form(NEXT,0,t,0)));
  case RELEASE:
    return new_form(AND,0,expand(t->right),new_form(OR,0,expand(t->left),new_form(NEXT,0,t,0)));
  }
  fatal_error("expand: end reached");
}



form* simplify(form* f){
  form* left;
  form* right;
  if(f==NULL)
    return f;
  if(idcmp(f->id,true->id))
    return true;
  if(idcmp(f->id,false->id))
    return false;  
  left=simplify(f->left);
  right=simplify(f->right);
  switch(f->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
    return f;
  case NEXT:
    return new_form(NEXT,0,left,0);
  case AND:
    if(idimply(left,right))
      return left;
    else if(idimply(right,left))
      return right;
    return new_form(AND,0,left,right);
  case OR:
    if(idimply(left,right))
      return right;
    else if(idimply(right,left))
      return left;
    return new_form(OR,0,left,right);
  case UNTIL:
    if(idimply(left,right))
      return right;
    return new_form(UNTIL,0,left,right);
  case RELEASE:
    if(idimply(right,left))
      return right;
    return new_form(RELEASE,0,left,right);
  }
  
  fatal_error("simplify: end reached");
}



form* propagate(form* f){
  form *left,*right;
  if(f==NULL)
    return f;
  left=propagate(f->left);
  right=propagate(f->right);
  switch(f->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
    return f;
  case NEXT:
    if(left->type==TRUE || left->type==FALSE)
      return left;
    else 
      return new_form(NEXT,0,left,0);
  case AND:
    if(left->type==FALSE || right->type==TRUE)
      return left;
    else if(left->type==TRUE || right->type==FALSE)
      return right;
    return new_form(AND,0,left,right);
  case OR:
    if(left->type==FALSE || right->type==TRUE)
      return right;
    else if(left->type==TRUE || right->type==FALSE)
      return left;
    return new_form(OR,0,left,right);
  case UNTIL:
    if(left->type==FALSE || right->type==TRUE)
      return right;
    return new_form(UNTIL,0,left,right);
  case RELEASE:
    if(left->type==TRUE || right->type==FALSE)
      return right;
    return new_form(RELEASE,0,left,right);
  }
  
  fatal_error("propagate: end reached");
}





form* assign(int prop,int value,form* f){
  switch(f->type){
  case TRUE:
  case FALSE:
    return f;
  case ATOM:
    if(f->prop!=prop)
      return f;
    else if(value)
      return true;
    else 
      return false;
  case NOT:
    if(f->prop!=prop)
      return f;
    else if(value)
      return false;
    else 
      return true;
  case NEXT:
    return f;
  case AND:
    return new_form(AND,0,assign(prop,value,f->left),assign(prop,value,f->right));
  case OR:
    return new_form(OR,0,assign(prop,value,f->left),assign(prop,value,f->right));
  }
  
  fatal_error("assign: end reached");
}

form* Xfactoring(form* f,int type){
  form *f1,*f2;
  if(f->type==NEXT)
    return f;
  f1=Xfactoring(f->left,type);
  f2=Xfactoring(f->right,type);
  if(f->type==type)
    return new_form_withid(NEXT,0,new_form(type,0,f1->left,f2->left),0,f->id);
  else
    return new_form(f->type,0,f1,f2);
}


void print_form_prefix(form* f,FILE* output){
switch(f->type){
  case TRUE:
    fprintf(output,"t ");
    break;
  case FALSE:
    fprintf(output,"f ");
    break;
  case ATOM:
    print_tree_prefix(props[f->prop],output);
    break;
  case NOT:
    fprintf(output,"! ");
    print_tree_prefix(props[f->prop],output);
    break;
  case NEXT:
    fprintf(output,"X ");
    print_form_prefix(f->left,output);
    break;
  case AND:
    fprintf(output,"& ");
    print_form_prefix(f->left,output);
    print_form_prefix(f->right,output);
    break;
  case OR:
    fprintf(output,"| ");
    print_form_prefix(f->left,output);
    print_form_prefix(f->right,output);
    break;
  case UNTIL:
    fprintf(output,"U ");
    print_form_prefix(f->left,output);
    print_form_prefix(f->right,output);
    break;
  case RELEASE:
    fprintf(output,"R ");
    print_form_prefix(f->left,output);
    print_form_prefix(f->right,output);
    break;
  }
  
}  


void print_form_spin(form* f,FILE* output){
switch(f->type){
  case TRUE:
    fprintf(output,"true");
    break;
  case FALSE:
    fprintf(output,"false");
    break;
  case ATOM:
    print_tree_spin(props[f->prop],output);
    break;
  case NOT:
    fprintf(output,"!(");
    print_tree_spin(props[f->prop],output);
    fprintf(output,")");
    break;
  case NEXT:
    fprintf(output,"X(");
    print_form_spin(f->left,output);
    fprintf(output,")");
    break;
  case AND:
    fprintf(output,"(");
    print_form_spin(f->left,output);
    fprintf(output,") && (");
    print_form_spin(f->right,output);
    fprintf(output,")");
    break;
  case OR:
    fprintf(output,"(");
    print_form_spin(f->left,output);
    fprintf(output,") || (");
    print_form_spin(f->right,output);
    fprintf(output,")");
    break;
  case UNTIL:
    fprintf(output,"(");
    print_form_spin(f->left,output);
    fprintf(output,") U (");
    print_form_spin(f->right,output);
    fprintf(output,")");
    break;
  case RELEASE:
    fprintf(output,"(");
    print_form_spin(f->left,output);
    fprintf(output,") V (");
    print_form_spin(f->right,output);
    fprintf(output,")");
    break;
  }
  
}  


void print_form_nusmv(form* f,FILE* output){
switch(f->type){
  case TRUE:
    fprintf(output,"TRUE");
    break;
  case FALSE:
    fprintf(output,"FALSE");
    break;
  case ATOM:
    print_tree_nusmv(props[f->prop],output);
    break;
  case NOT:
    fprintf(output,"!(");
    print_tree_nusmv(props[f->prop],output);
    fprintf(output,")");
    break;
  case NEXT:
    fprintf(output,"X(");
    print_form_nusmv(f->left,output);
    fprintf(output,")");
    break;
  case AND:
    fprintf(output,"(");
    print_form_nusmv(f->left,output);
    fprintf(output,") & (");
    print_form_nusmv(f->right,output);
    fprintf(output,")");
    break;
  case OR:
    fprintf(output,"(");
    print_form_nusmv(f->left,output);
    fprintf(output,") | (");
    print_form_nusmv(f->right,output);
    fprintf(output,")");
    break;
  case UNTIL:
    fprintf(output,"(");
    print_form_nusmv(f->left,output);
    fprintf(output,") U (");
    print_form_nusmv(f->right,output);
    fprintf(output,")");
    break;
  case RELEASE:
    fprintf(output,"(");
    print_form_nusmv(f->left,output);
    fprintf(output,") V (");
    print_form_nusmv(f->right,output);
    fprintf(output,")");
    break;
  }
  
}  


void print_form_wring_infix(form* f,FILE* output){
  switch(f->type){
  case TRUE:
    fprintf(output,"(TRUE)");
    break;
  case FALSE:
    fprintf(output,"(FALSE)");
    break;
  case ATOM:
    fprintf(output,"(p%d=1)",f->prop);
    break;
  case NOT:
    fprintf(output,"(p%d=0)",f->prop);
    break;
  case NEXT:
    fprintf(output,"(X");
    print_form_wring_infix(f->left,output);
    fprintf(output,")");
    break;
  case AND:
    fprintf(output,"(");
    print_form_wring_infix(f->left,output);
    fprintf(output,"*");
    print_form_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  case OR:
    fprintf(output,"(");
    print_form_wring_infix(f->left,output);
    fprintf(output,"+");
    print_form_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  case UNTIL:
    fprintf(output,"(");
    print_form_wring_infix(f->left,output);
    fprintf(output,"U");
    print_form_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  case RELEASE:
    fprintf(output,"(");
    print_form_wring_infix(f->left,output);
    fprintf(output,"R");
    print_form_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  }
  
}  


int print_form_wring(form* f,FILE* output){
  int l,r;
  switch(f->type){
  case TRUE:
    return 0;
  case ATOM:
    fprintf(output,"p%d=1",f->prop);
    return 1;
  case NOT:
    fprintf(output,"p%d=0",f->prop);
    return 1;
  case NEXT:
    print_form_wring_infix(f,output);
    return 1;
  case AND:
    l=print_form_wring(f->left,output);
    if(l)
      fprintf(output,",");
    r=print_form_wring(f->right,output);
    return l+r;
  }
  fatal_error("print_form_wring: end reached");
  
}  

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [It creates a new id for the formula t.]

  Description [First, it tries to derive the id of t from the
    ids of existing formulas. If it cannot, it produces a new id based
    on the ids of subformulas. ]

******************************************************************************/
void creaid(form* t){
  list* tmp=formulas;
  switch(t->type){
  case TRUE:
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==TRUE){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==FALSE){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case FALSE:
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==FALSE){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==TRUE){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case ATOM:
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==ATOM && t->prop==t2->prop){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==NOT && t->prop==t2->prop){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case NOT:
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==NOT && t->prop==t2->prop){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==ATOM && t->prop==t2->prop){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case NEXT:
    if(t->left->id!=NULL){
      if((t->left->type==AND || t->left->type==OR)){
	idtype i1=NULL,i2=NULL;
	int tf1=0,tf2=0;
	form* t3;
	while(tmp!=NULL && (i1==NULL || i2==NULL)){
	  t3=tmp->el;
	  if (t3->id != NULL) {
	    if(t3->type==NEXT && i1==NULL && idcmp(t3->left->id,t->left->left->id))
	      i1=t3->id;
	    if(t3->type==NEXT && i2==NULL && idcmp(t3->left->id,t->left->right->id))
	      i2=t3->id;
	    if(t3->type==NEXT && i1==NULL && idcmpnot(t3->left->id,t->left->left->id)){
	      i1=idnot(t3->id);
	      tf1=1;
	    }
	    if(t3->type==NEXT && i2==NULL && idcmpnot(t3->left->id,t->left->right->id)){
	      i2=idnot(t3->id);
	      tf2=1;
	    }
	  }
	  tmp=tmp->next;
	}
	if(i1!=NULL && i2!=NULL){
	  if(t->left->type==AND)
	    t->id=idand(i1,i2);
	  else
	    t->id=idor(i1,i2);
	}
	if(tf1)
	  free(i1);
	if(tf2)
	  free(i2);
      }
      if(t->id==NULL && (idcmp(t->left->id,false->id) || idcmp(t->left->id,true->id)))
	t->id=t->left->id;
      tmp=formulas;
      while(t->id==NULL && tmp!=NULL){
	form* t2=(form*)tmp->el;
	if (t2->id != NULL) {
	  if(t2->type==NEXT && idcmp(t->left->id,t2->left->id))
	    t->id=t2->id;
	  else if(t2->type==NEXT && idcmpnot(t->left->id,t2->left->id))
	    t->id=idnot(t2->id);
	  else
	    tmp=tmp->next;
	} else
	  tmp=tmp->next;
      } 
    }
    break;
  case AND:
    if(idimply(t->left,t->right)){
      t->id=t->left->id;
      tmp=NULL;
    } else if(idimply(t->right,t->left)){
      t->id=t->right->id;
      tmp=NULL;
    }   
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==AND && idcmp(t->left->id,t2->left->id) && idcmp(t->right->id,t2->right->id)){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==OR && idcmpnot(t->left->id,t2->left->id) && idcmpnot(t->right->id,t2->right->id)){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case OR:
    if(idimply(t->left,t->right)){
      t->id=t->right->id;
      tmp=NULL;
    } else if(idimply(t->right,t->left)){
      t->id=t->left->id;
      tmp=NULL;
    }    
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==OR && idcmp(t->left->id,t2->left->id) && idcmp(t->right->id,t2->right->id)){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==AND && idcmpnot(t->left->id,t2->left->id) && idcmpnot(t->right->id,t2->right->id)){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case UNTIL:
    if(idimply(t->left,t->right)){
      t->id=t->right->id;
      tmp=NULL;
    }
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==UNTIL && idcmp(t->left->id,t2->left->id) && idcmp(t->right->id,t2->right->id)){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==RELEASE && idcmpnot(t->left->id,t2->left->id) && idcmpnot(t->right->id,t2->right->id)){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  case RELEASE:
    if(idimply(t->right,t->left)){
      t->id=t->right->id;
      tmp=NULL;
    }
    while(tmp!=NULL){
      form* t2=(form*)tmp->el;
      if (t2->id != NULL) {
	if(t2->type==RELEASE && idcmp(t->left->id,t2->left->id) && idcmp(t->right->id,t2->right->id)){
	  t->id=t2->id;
	  tmp=NULL;
	} else if(t2->type==UNTIL && idcmpnot(t->left->id,t2->left->id) && idcmpnot(t->right->id,t2->right->id)){
	  t->id=idnot(t2->id);
	  tmp=NULL;
	} else
	  tmp=tmp->next;
      } else
	tmp=tmp->next;
    }
    break;
  }

  if(t->id==NULL){
    new_id(t);
    if(t->type==UNTIL || t->type==RELEASE)
      if(idcmp(t->id,t->right->id)){
	form* t3=new_form(NEXT,0,t,0);
	idtype oldid=t3->id;
	t3->id=NULL;
	tmp=formulas;
	while(t3->id==NULL && tmp!=NULL){
	  form* t2=(form*)tmp->el;
	  if (t2->id != NULL) {
	    if(t2->type==NEXT && idcmp(t->id,t2->left->id) && t2!=t3)
	      t3->id=t2->id;
	    else if(t2->type==NEXT && idcmpnot(t->id,t2->left->id))
	      t3->id=idnot(t2->id);
	  }
	  tmp=tmp->next;
	}
	if(t3->id!=NULL)
	  free(oldid);
	else
	  t3->id=oldid;
      }
  }
}
