/**CFile***********************************************************************

  FileName    [id.c]

  Package     [id]

  Synopsis    [Formalas' id manipulation]

  Description [This file provides the function to create and
  manipulate the ids' of formulas. Ids are used for formulas
  entailment. ]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.9]

******************************************************************************/

#include "form.h"
#include "formInt.h"
#include "prop.h"
#include "modella.y.tab.h" 
#include "main.h"
#include "tree.h"
#include "utils.h"

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
/* counter of ids created for elementary subformulas */
static size_t id_lit_counter;
/* upper bound for id_lit_counter */
static size_t max_lit; 


/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static idtype id_alloc();
static void initialize_lit();
static void initialize_formid();
static void initialize_formid_recur(list *);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [It initializes the id global variables.]

******************************************************************************/
void initialize_id(){
  /* number of bit in idtokentype */ 
  const int tok_bits = 8 * sizeof(idtokentype);
  
  id_lit_counter=0;
  max_lit=0;
  fair_number=0;
  initialize_lit();

  /* the number of bit for every id is 2^max_lit */
  id_size = modella_pow2(max_lit)/tok_bits;
  
  /* when max_lit is small, we need at least one byte */
  if(id_size == 0)
    id_size=1;

  initialize_formid();
}
  
/**Function********************************************************************

  Synopsis    [It creates a new id for the formula f.]

  Description [If f is elementary the id is a sequence of the form
  11...100...011...100...0...; i.e., there are 2n blocks alternatively
  of 1's and 0's. The length of the block is given by 2^id_lit, where
  id_lit is equal to f->prop in the case of literals, while it is
  assigned to a unique value (given by id_lit_counter) in the case of
  X-formulas.]

  SideEffects [If f is a X-formula, id_lit_counter is increased]

******************************************************************************/
void new_id(form* f){
  int i,j;

  /* number of bit in idtokentype */ 
  const int tok_bits = 8 * sizeof(idtokentype);
  /* total number of bits in each id */
  const int id_bits = id_size*tok_bits;

  /* number that identifies an elementary formula: it is equal to prop
     for literals, while it is assigned to id_lit_counter in the case
     of next formulas */
  int id_lit;
  /* length of each block (see function description) */
  int block_length;

  switch(f->type){
  case ATOM:
    id_lit = f->prop;
    modella_assert(id_lit < prop_number);
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=0;
   
    block_length = modella_pow2(id_lit);

    for (i=0; i<id_bits; i++){
      /* the i-th bit is one */
      int one_block = (i%(2*block_length) < block_length);
      if (one_block) {
	/* the i-th bit is in the j-th token */
	int j = i / tok_bits;
	/* the i-th bit is the k-th bit of the j-th token */
	int k = i % tok_bits;
	f->id[j] += modella_pow2(k);
      } /* else the i-th bit is zero and I do nothing */
    }
    
    break;
  case NOT:
    id_lit = f->prop;
    modella_assert(id_lit < prop_number);
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=0;
   
    block_length = modella_pow2(id_lit);

    for (i=0; i<id_bits; i++){
      /* the i-th bit is one */
      int one_block = (i%(2*block_length) >= block_length);
      if (one_block) {
	/* the i-th bit is in the j-th token */
	int j = i / tok_bits;
	/* the i-th bit is the k-th bit of the j-th token */
	int k = i % tok_bits;
	f->id[j] += modella_pow2(k);
      } /* else the i-th bit is zero and I do nothing */
    }
    
    break;
  case TRUE:
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=-1;
    break;
  case FALSE:
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=0;
    break;
  case NEXT:
    id_lit = id_lit_counter++;
    modella_assert(id_lit < max_lit);
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=0;
    
    block_length = modella_pow2(id_lit);

    for (i=0; i<id_bits; i++){
      /* the i-th bit is one */
      int one_block = (i%(2*block_length) < block_length);
      if (one_block) {
	/* the i-th bit is in the j-th token */
	int j = i / tok_bits;
	/* the i-th bit is the k-th bit of the j-th token */
	int k = i % tok_bits;
	f->id[j] += modella_pow2(k);
      } /* else the i-th bit is zero and I do nothing */
    }
    
    break;
  case AND:
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=f->left->id[j]&f->right->id[j];
    break;
  case OR:
    f->id=id_alloc();
    for(j=0;j<id_size;j++)
      f->id[j]=f->left->id[j]|f->right->id[j];
    break;
  case RELEASE:
    {
      form* t = new_form(NEXT,0,f,0);
      f->id=id_alloc();
      for(j=0;j<id_size;j++)
	f->id[j]=f->right->id[j]&(f->left->id[j]|t->id[j]);
      break;
    }
  case UNTIL:
    {
      form* t = new_form(NEXT,0,f,0);
      f->id=id_alloc();
      for(j=0;j<id_size;j++)
	f->id[j]=f->right->id[j]|(f->left->id[j]&t->id[j]);
      break;
    }
  }
}
 
/**Function********************************************************************

  Synopsis    [It returns the negation of id]

******************************************************************************/
idtype idnot(idtype id){
  int j;
  idtype id2=id_alloc();
  for(j=0;j<id_size;j++)
    id2[j]=~id[j];
  return id2;
}

/**Function********************************************************************

  Synopsis    [It returns the conjunction of id1 and id2]

******************************************************************************/
idtype idand(idtype id1,idtype id2){
  int j;
  idtype id3=id_alloc();
  for(j=0;j<id_size;j++)
    id3[j]=id1[j]&id2[j];
  return id3;
}

/**Function********************************************************************

  Synopsis    [It returns the disjunction of id1 and id2]

******************************************************************************/
idtype idor(idtype id1,idtype id2){
  int j;
  idtype id3=id_alloc();
  for(j=0;j<id_size;j++)
    id3[j]=id1[j]|id2[j];
  return id3;
}

/**Function********************************************************************

  Synopsis    [It returns 1 if id1 and id2 are equal.]

******************************************************************************/
int idcmp(idtype id1,idtype id2){
  int j;
  for(j=0;j<id_size;j++)
    if(id1[j]!=id2[j])
      return 0;
  return 1;
}

/**Function********************************************************************

  Synopsis    [It returns 1 if id1 and id2 are not equal.]

******************************************************************************/
int idcmpnot(idtype id1,idtype id2){
  int j;
  for(j=0;j<id_size;j++)
    if(id1[j]!=~id2[j])
      return 0;
  return 1;
}

/**Function********************************************************************

  Synopsis    [It copies id2 into id1.]

******************************************************************************/
void idcpy(idtype id1,idtype id2){
  int j;
  for(j=0;j<id_size;j++)
    id1[j]=id2[j];
}

/**Function********************************************************************

  Synopsis    [It returns 1 if id1 implies id2.]

******************************************************************************/
int idimply(form* left,form* right){
  int j;
  for(j=0;j<id_size;j++)
    if(((~left->id[j])|right->id[j])!=-1)
      return 0;
  return 1;
}
  
  
/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
    
/**Function********************************************************************

  Synopsis           [It creates a new id.]

******************************************************************************/
static idtype id_alloc(){
  idtype id = (idtype)mymalloc(id_size*sizeof(idtokentype));
}

/**Function********************************************************************

  Synopsis    [It inititializes id global variables.]

  Description [We scan the list of formulas to find the number of elementary
  formulas. ]

  SideEffects [max_lit is increased by the number of atoms in t]

  SeeAlso     [initialize_id]

******************************************************************************/
static void initialize_lit(){
  list* tmp=formulas;
  while(tmp!=NULL){
    form* t=(form*)tmp->el;
    switch(t->type){
    case TRUE:
    case FALSE:
    case OR:
    case AND:
      break;
    case ATOM:
    case NOT:
      id_lit_counter++;
      max_lit++;
      break;
    case NEXT:
      max_lit++;
      break;
    case UNTIL:
      fair_number++;
      max_lit++;
      break;
    case RELEASE:
      max_lit++;
      break;
    default:
      fatal_error("initialize_lit: unexpected switch case");
    }
    tmp = tmp->next;
  }
}

/**Function********************************************************************

  Synopsis    [It inititializes id of created formulas.]

  Description [We scan the list of formulas and create a new id for
  each of them. ]

  SideEffects [The id of form in formulas is initiailized.]

  SeeAlso     [initialize_id]

******************************************************************************/
static void initialize_formid(){
  initialize_formid_recur(formulas);
}

/**Function********************************************************************

  Synopsis    []

******************************************************************************/
static void initialize_formid_recur(list *fp){
  if (fp != NULL) {
    form *f = (form*) fp->el;;
    initialize_formid_recur(fp->next);
    
    /* if f is an U or a V formula, 
       then we first create the id the Xf. */
    if (f->type == UNTIL || f->type == RELEASE){
      form* next_f = new_form(NEXT,0,f,0);
      creaid(next_f);
    }

    if(f->id == NULL)
      creaid(f);
  }
}

