/**CFile***********************************************************************

  FileName    [main.c]

  Package     [main]

  Synopsis    [It contains the main function.]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.9]

******************************************************************************/

#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<errno.h>
#include<unistd.h>

#include "main.h"
#include "prop.h"
#include "struct.h"
#include "modella.y.tab.h" 
#include "tree.h"
#include "form.h"
#include "automa.h"
#include "GL02.h"
#include "EWS02.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/
const char* VERSION="1.5.9";

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
extern FILE* yyin;

int yyparse();


  

void usage(){
  fprintf(stderr,"modella version %s");
  fprintf(stderr,"\
usage: modella file_input file_output\n\
  -r<N> set the rewriting rules mode (0:disabled,1:SB,2:EH,12:both,default:0)\n\
  -g enable the optimizations of GL02\n\
  -e enable the optimizations of EWS02\n\
  -o<N> set \"merging states\" optimization (1:true,0:false,default:1)\n\
  -w print the automaton in the syntax of Wring (-e and -g disabled)\n\
  -s use Spin's syntax, print a never claim\n\
  -n use NuSMV's syntax, print a GBA in XML (-e and -g disabled)\n\
  -m use NuSMV's syntax, print a GBA in SMV (-e and -g disabled)\n\
  -f in combination with -n o -m, produce a weak automaton\n\
  -P write to automa.dot the automaton in the syntax of dot\n\
");
    exit(-1); 
}

int main(int argc, char **argv){
  FILE* input;
  FILE* output;

  GBA* b;
  LBA* lb;

  form* formula;

  int dot=0;
  int gl02=0;
  int ews02=0;
  enum {lbtt,wring,spin,smv,xmlbuechi} syntax_output=lbtt;
  int l2s=0;
  int writeversion=0;
  syntax=lbtt;
  fair_number=0;
  verbosity=0;
  rewrules=0;
  optim=1;
  weak=0;
  


  for(--argc, argv++; argc; argc--, argv++){
    if(argv[0][0]=='-'){
      argv[0]++; 
      switch (argv[0][0]) {
      case 'v':
	argv[0]++;
	verbosity=atoi(argv[0]);
	break;
      case 'r':
	argv[0]++;
	rewrules=atoi(argv[0]);
	break;
      case 'o':
	argv[0]++;
	optim=atoi(argv[0]);
	break;
      case 'w':
	syntax_output=wring;
	break;
      case 's':
	syntax=spin;
	syntax_output=spin;
	break;
      case 'n':
	syntax=nusmv;
	syntax_output=xmlbuechi;
	break;
      case 'm':
	syntax=nusmv;
	syntax_output=smv;
	break;
      case 'f':
	weak=1;
	//case 'd':
	//degen=1;
	//break;
      case 'P':
	dot=1;
	break;
      case 'g':
	gl02=1;
	break;
      case 'e':
	ews02=1;
	break;
      case 'l':
	l2s=1;
	break;
      case 'V':
	writeversion=1;
	break;
      default:
	usage();
      }
    } else
      break;
  }
  
  /* if no input is specified, use stdin */
  if(argc>0) {  
    input=fopen(argv[0],"r");
    if(input==NULL){
      perror("open input failed");
      exit(-1);
    }
  } else
    input = stdin;

  /* if no output is specified, use stdout */
  if(argc>1) {
    output=fopen(argv[1],"w");
    if(output==NULL){
      perror("open output failed");
      exit(-1);
    }
  } else
    output = stdout;
  
  yyin=input;
  if(yyparse())
    return -1;
      
  parsing_tree = NNF(parsing_tree);

  if(rewrules!=0)
    parsing_tree = semplifytree(parsing_tree, rewrules);
  
  /* initialize the prop package to identify the set of propositions */
  initialize_prop();
  /* initialize the form package */
  initialize_form();
      
  /* fill the list of subformulas of the input formula */
  formula=tree2form(parsing_tree);
      
  /* initialize the id package */
  initialize_id();
  /* initialize the number of fair conditions */
  initialize_faircond();


  /* convert the formula into a GBA */
  b=new_GBA(formula);



     
  if(syntax_output==wring)
    print_GBAw(b,output);
  else if(syntax_output==xmlbuechi) 
    print_GBAn(b,output);
  else if(syntax_output==smv) 
    print_GBAm(b,output);
  else {
    if(gl02)
      lb=TLGBAtoLBA(GBAtoTLGBA(b));
    else
      lb=BAtoLBA(GBAtoBA(b));
	
	
    lb=prune(lb);
	
	
    if(ews02)
      lb=sim_reduce(lb);
	
    lb=prune(lb);
	
    if(syntax_output==spin){
      if(l2s)
	print_LBAs_l2s(lb,output);
      else
	print_LBAs(lb,output);
    } else
      print_LBA(lb,output);
	
    if(verbosity>1)
      print_LBA2(lb,stdout);
	
    if(dot)
      LBAtodot(lb);
  }	
      
  fflush(output);
  fflush(stderr);
  fclose(input);
  fclose(output);
      
  

  exit(0);
}

