/**CFile***********************************************************************

  FileName    [prop.c]

  Package     [prop]

  Synopsis    [Formalas' propositions manipulation]

  Description [This file provides the function to identify and store
  the propositions that occur in the formulas.]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.9]

******************************************************************************/

#include "tree.h"
#include "prop.h"
#include "main.h"
#include "modella.y.tab.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int only_occurrence(tree* t1,tree *t2);
static int only_propoccurrence(tree* t);
static void initialize_prop_recur(tree* t);
static tree* update(tree* t);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [It inititializes props according to the propositions
  in t.]

  SideEffects [props is initialized.]

  SeeAlso     [initialize_prop_recur]

******************************************************************************/
void initialize_prop(){
  prop_number=0;
  initialize_prop_recur(parsing_tree);
  parsing_tree = update(parsing_tree);
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
    
/**Function********************************************************************

  Synopsis           [It checks if t1 is the only occurrence of the
  corresponding atom in t2]

  Description        [t1 must be a literal. The function checks if
  there are other occurrence of t1 or its negation in t2.]

  SeeAlso            [only_propoccurence]

******************************************************************************/
static int only_occurrence(tree* t1,tree *t2){
  modella_assert((t1->type==ATOM || t1->type==NOT));

  if(t2==NULL)
    return 1;
  else if(t1==t2)
    return 1; // We are looking an occurrence different from t1
  else if((t2->type==ATOM || t2->type==NOT) && !strcmp(t1->name,t2->name))
    return 0;
  else return only_occurrence(t1,t2->left) &&
	 only_occurrence(t1,t2->right);
}

/**Function********************************************************************

  Synopsis           [It checks if t is a boolean combination of atoms
  that occurr only once]

******************************************************************************/
static int only_propoccurrence(tree* t){
  if(t==NULL)
    return 1;
  else if(t->type==ATOM || t->type==NOT)
    return only_occurrence(t,parsing_tree);
  else if(t->type==NEXT || t->type==UNTIL || t->type==RELEASE)
    return 0;
  else return only_propoccurrence(t->left) &&
	 only_propoccurrence(t->right);
}

/**Function********************************************************************

  Synopsis    [It inititializes props according to the propositions in t.]

  Description [For every atom in t, the function allocs a position in
  props. It also checks if t is a booleand combination of literals:
  if it is, t is considered as an atom. This last change is performed
  in order to optimize the translation when the input formula contains
  long boolean expressions as leaves of LTL. These cases make the
  representation blow up, if we consider every literal as an atom.]
   
  SideEffects [props is updated.]

  SeeAlso     [initialize_id]

******************************************************************************/
static void initialize_prop_recur(tree* t){
  /* new_props is the new version of props*/
  tree **new_props;

  int i;
  switch(t->type){
  case TRUE:
  case FALSE:
    return;
  case ATOM:
    /* If t is already in props, return */
    for(i=0;i<prop_number;i++)
      if(props[i]->type==ATOM && !strcmp(props[i]->name,t->name))
	return;

    /* else add t to props */
    new_props=(tree**)mymalloc((prop_number+1)*sizeof(tree*));
    for(i=0;i<prop_number;i++)
      new_props[i]=props[i];
    new_props[prop_number]=new_tree(ATOM,t->name,0,0);
    if(prop_number>0)
      myfree(props);
    props=new_props;
    prop_number++;
    return;
  case NOT:
    /* If t is already in props, return */
    for(i=0;i<prop_number;i++)
      if(props[i]->type==ATOM && !strcmp(props[i]->name,t->name))
	return;

    /* else add t to props */
    new_props=(tree**)mymalloc((prop_number+1)*sizeof(tree*));
    for(i=0;i<prop_number;i++)
      new_props[i]=props[i];
    new_props[prop_number]=new_tree(ATOM,t->name,0,0);
    if(prop_number>0)
      myfree(props);
    props=new_props;
    prop_number++;
    return;
  case OR:
  case AND:
    if(only_propoccurrence(t)){
      for(i=0;i<prop_number;i++)
	if(treecmp(props[i],t))
	  return;
      new_props=(tree**)mymalloc((prop_number+1)*sizeof(tree*));
      for(i=0;i<prop_number;i++)
	new_props[i]=props[i];
      new_props[prop_number]=dup_tree(t);
      if(prop_number>0)
	myfree(props);
      prop_number++;
      props=new_props;
      return;
    } else {
      initialize_prop_recur(t->left);
      initialize_prop_recur(t->right);
      return;
    }
  case NEXT:
    initialize_prop_recur(t->left);
    return;
  case UNTIL:
  case RELEASE:
    initialize_prop_recur(t->left);
    initialize_prop_recur(t->right);
    return;
  }
  fatal_error("initialize_prop_recur: end reached");
}

/**Function*******************************************************************

  Synopsis    [It updates the parsing tree. ]

  Description [It updates the parsing tree according to the new set of
  propositions. ]

******************************************************************************/
static tree* update(tree* t){
  int i;
  tree* tmp;
  switch(t->type){
  case TRUE:
  case FALSE:
    return t;
  case ATOM:
    for(i=0;i<prop_number;i++)
      if(props[i]->type==ATOM && !strcmp(props[i]->name,t->name)){
	tmp=new_tree(ATOM,0,0,0);
	tmp->prop=i;
	myfree(t);
	return tmp;
      }
    return t;
  case NOT:
    for(i=0;i<prop_number;i++)
      if(props[i]->type==ATOM && !strcmp(props[i]->name,t->name)){
	tmp=new_tree(NOT,0,0,0);
	tmp->prop=i;
	myfree(t);
	return tmp;
      }
    return t;
  case OR:
  case AND:
    for(i=0;i<prop_number;i++)
      if(treecmp(props[i],t)){
	tmp=new_tree(ATOM,0,0,0);
	tmp->prop=i;
	myfree(t);
	return tmp;
      }
    t->left=update(t->left);
    t->right=update(t->right);
    return t;
  case NEXT:
    t->left=update(t->left);
    return t;
  case UNTIL:
  case RELEASE:
    t->left=update(t->left);
    t->right=update(t->right);
    return t;
  }
  fprintf(stderr,"fatal error in update: end reached\n" );
  exit(-1);
}
