/**CFile***********************************************************************

  FileName    [tree.c]

  Package     [tree]

  Synopsis    [Tree manipulation]

  Description [This file provides the function to create and
  manipulate the parsing tree of a formula.]

  Author      [Stefano Tonetta]

  Copyright   [This file is part of MoDeLLa.
  Copyright (C) 2003-2006 by Stefano Tonetta and Roberto Sebastiani.]

  Version     [1.5.8.2]

******************************************************************************/

#include "tree.h"
#include "main.h"
#include "modella.y.tab.h"
#include "form.h"
#include "rew.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static tree* NNF_recur(tree* t,int flag);
static form* tree2form_recur(tree* t);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [It creates a new tree.]

******************************************************************************/
tree* new_tree(int type,char* name,tree* left,tree* right){
  tree* t=(tree*)mymalloc(sizeof(tree));
  t->type=type;
  t->name=name;
  t->left=left;
  t->right=right;
  return t;
}


/**Function********************************************************************

  Synopsis    [It creates a copy of a tree.]

******************************************************************************/
tree* dup_tree(tree* t){
  if(t==NULL)
    return t;
  else {
    return
    new_tree(t->type,t->name,dup_tree(t->left),dup_tree(t->right));
  }
}

/**Function********************************************************************

  Synopsis    [It destroys a tree.]

******************************************************************************/
void delete_tree(tree* t){
  switch(t->type){
  case TRUE:
  case FALSE:
    break;
  case ATOM:
  case NOT:
    break;
  case NEXT:
    delete_tree(t->left);
    break;
  case AND:
  case OR:
  case UNTIL:
  case RELEASE:
    delete_tree(t->left);
    delete_tree(t->right);
    break;
  }
  myfree(t);
}

/**Function********************************************************************

  Synopsis    [It converts the tree in NNF.]

******************************************************************************/
tree* NNF(tree* t){
  return NNF_recur(t,1);
}

/**Function********************************************************************

  Synopsis    [It converts a tree into a formula.]

******************************************************************************/
form* tree2form(tree* t){
  return tree2form_recur(t);
}

/**Function********************************************************************

  Synopsis    [It prints the contents of a tree.]

******************************************************************************/
void print_tree(tree* t){
  switch(t->type){
  case TRUE:
    fprintf(stderr,"t ");
    break;
  case FALSE:
    fprintf(stderr,"f ");
    break;
  case ATOM:
    fprintf(stderr,"%s ",t->name);
    break;
  case NOT:
    if(t->left==NULL)
      fprintf(stderr,"!%s ",t->name);
    else {
      fprintf(stderr,"! ");
      print_tree(t->left);
    }
    break;
  case NEXT:
    fprintf(stderr,"X ");
    print_tree(t->left);
    break;
  case AND:
    fprintf(stderr,"& ");
    print_tree(t->left);
    print_tree(t->right);
    break;
  case OR:
    fprintf(stderr,"| ");
    print_tree(t->left);
    print_tree(t->right);
    break;
  case UNTIL:
    fprintf(stderr,"U ");
    print_tree(t->left);
    print_tree(t->right);
    break;
  case RELEASE:
    fprintf(stderr,"V ");
    print_tree(t->left);
    print_tree(t->right);
    break;
  }
}

/**Function********************************************************************

  Synopsis    [It simplifies a tree according to some rewriting rules.]

******************************************************************************/
tree* semplifytree(tree *t, int rr){
  if(rr==1)
    return rewrite1(t);
  else if(rr==2)
    return rewrite2(t);
  else 
    return rewrite12(t);
}

/**Function********************************************************************

  Synopsis    [It returns the dual tree.]

******************************************************************************/
tree* dual(tree* t){
  switch(t->type){
  case TRUE:
    return new_tree(FALSE,0,0,0);
  case FALSE:
    return new_tree(TRUE,0,0,0);
  case ATOM:
    return new_tree(NOT,t->name,0,0);
  case NOT:
    return new_tree(ATOM,t->name,0,0);
  case NEXT:
    return new_tree(NEXT,0,dual(t->left),0);
  case AND:
    return new_tree(OR,0,dual(t->left),dual(t->right));
  case OR:
    return new_tree(AND,0,dual(t->left),dual(t->right));
  case UNTIL:
    return new_tree(RELEASE,0,dual(t->left),dual(t->right));
  case RELEASE:
    return new_tree(UNTIL,0,dual(t->left),dual(t->right));
  }
  
}

int treecmp(tree* t1,tree* t2){
  if(t1==NULL || t2==NULL)
    return (t1==t2);
  else if(t1->name==NULL || t2->name==NULL)
    return (t1->type==t2->type && t1->name==t2->name && treecmp(t1->left,t2->left) && treecmp(t1->right,t2->right));
  else
    return (t1->type==t2->type && !strcmp(t1->name,t2->name) && treecmp(t1->left,t2->left) && treecmp(t1->right,t2->right));
}

int aredual(tree* t1,tree* t2){
  switch(t1->type){
  case TRUE:
    return (t2->type==FALSE);
  case FALSE:
    return (t2->type==TRUE);
  case ATOM:
    return (t2->type==NOT && !strcmp(t1->name,t2->name));
  case NOT:
    return (t2->type==ATOM && !strcmp(t1->name,t2->name));
  case NEXT:
    return (t2->type==NEXT && aredual(t1->left,t2->left));
  case AND:
    return (t2->type==OR && aredual(t1->left,t2->left) && aredual(t1->right,t2->right));
  case OR:
    return (t2->type==AND && aredual(t1->left,t2->left) && aredual(t1->right,t2->right));
  case UNTIL:
    return (t2->type==RELEASE && aredual(t1->left,t2->left) && aredual(t1->right,t2->right));
  case RELEASE:
    return (t2->type==UNTIL && aredual(t1->left,t2->left) && aredual(t1->right,t2->right));
  }
  
}





void print_tree_prefix(tree* f,FILE* output){
switch(f->type){
  case TRUE:
    fprintf(output,"t ");
    break;
  case FALSE:
    fprintf(output,"f ");
    break;
  case ATOM:
    fprintf(output,"%s ",f->name);
    break;
  case NOT:
    fprintf(output,"! %s ",f->name);
    break;
  case NEXT:
    fprintf(output,"X ");
    print_tree_prefix(f->left,output);
    break;
  case AND:
    fprintf(output,"& ");
    print_tree_prefix(f->left,output);
    print_tree_prefix(f->right,output);
    break;
  case OR:
    fprintf(output,"| ");
    print_tree_prefix(f->left,output);
    print_tree_prefix(f->right,output);
    break;
  case UNTIL:
    fprintf(output,"U ");
    print_tree_prefix(f->left,output);
    print_tree_prefix(f->right,output);
    break;
  case RELEASE:
    fprintf(output,"R ");
    print_tree_prefix(f->left,output);
    print_tree_prefix(f->right,output);
    break;
  }
  
}  


void print_tree_spin(tree* f,FILE* output){
switch(f->type){
  case TRUE:
    fprintf(output,"true");
    break;
  case FALSE:
    fprintf(output,"false");
    break;
  case ATOM:
    fprintf(output,"%s",f->name);
    break;
  case NOT:
    fprintf(output,"!(%s)",f->name);
    break;
  case NEXT:
    fprintf(output,"X(");
    print_tree_spin(f->left,output);
    fprintf(output,")");
    break;
  case AND:
    fprintf(output,"(");
    print_tree_spin(f->left,output);
    fprintf(output,") && (");
    print_tree_spin(f->right,output);
    fprintf(output,")");
    break;
  case OR:
    fprintf(output,"(");
    print_tree_spin(f->left,output);
    fprintf(output,") || (");
    print_tree_spin(f->right,output);
    fprintf(output,")");
    break;
  case UNTIL:
    fprintf(output,"(");
    print_tree_spin(f->left,output);
    fprintf(output,") U (");
    print_tree_spin(f->right,output);
    fprintf(output,")");
    break;
  case RELEASE:
    fprintf(output,"(");
    print_tree_spin(f->left,output);
    fprintf(output,") V (");
    print_tree_spin(f->right,output);
    fprintf(output,")");
    break;
  }
  
}  


void print_tree_nusmv(tree* f,FILE* output){
switch(f->type){
  case TRUE:
    fprintf(output,"TRUE");
    break;
  case FALSE:
    fprintf(output,"FALSE");
    break;
  case ATOM:
    fprintf(output,"%s",f->name);
    break;
  case NOT:
    fprintf(output,"!(%s)",f->name);
    break;
  case NEXT:
    fprintf(output,"X(");
    print_tree_nusmv(f->left,output);
    fprintf(output,")");
    break;
  case AND:
    fprintf(output,"(");
    print_tree_nusmv(f->left,output);
    fprintf(output,") & (");
    print_tree_nusmv(f->right,output);
    fprintf(output,")");
    break;
  case OR:
    fprintf(output,"(");
    print_tree_nusmv(f->left,output);
    fprintf(output,") | (");
    print_tree_nusmv(f->right,output);
    fprintf(output,")");
    break;
  case UNTIL:
    fprintf(output,"(");
    print_tree_nusmv(f->left,output);
    fprintf(output,") U (");
    print_tree_nusmv(f->right,output);
    fprintf(output,")");
    break;
  case RELEASE:
    fprintf(output,"(");
    print_tree_nusmv(f->left,output);
    fprintf(output,") V (");
    print_tree_nusmv(f->right,output);
    fprintf(output,")");
    break;
  }
  
}  


void print_tree_wring_infix(tree* f,FILE* output){
  switch(f->type){
  case TRUE:
    fprintf(output,"(TRUE)");
    break;
  case FALSE:
    fprintf(output,"(FALSE)");
    break;
  case ATOM:
    fprintf(output,"(p%d=1)",f->prop);
    break;
  case NOT:
    fprintf(output,"(p%d=0)",f->left->prop);
    break;
  case NEXT:
    fprintf(output,"(X");
    print_tree_wring_infix(f->left,output);
    fprintf(output,")");
    break;
  case AND:
    fprintf(output,"(");
    print_tree_wring_infix(f->left,output);
    fprintf(output,"*");
    print_tree_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  case OR:
    fprintf(output,"(");
    print_tree_wring_infix(f->left,output);
    fprintf(output,"+");
    print_tree_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  case UNTIL:
    fprintf(output,"(");
    print_tree_wring_infix(f->left,output);
    fprintf(output,"U");
    print_tree_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  case RELEASE:
    fprintf(output,"(");
    print_tree_wring_infix(f->left,output);
    fprintf(output,"R");
    print_tree_wring_infix(f->right,output);
    fprintf(output,")");
    break;
  }
  
}  


int print_tree_wring(tree* f,FILE* output){
  int l,r;
  switch(f->type){
  case TRUE:
    return 0;
  case ATOM:
    fprintf(output,"%s=1",f->name);
    return 1;
  case NOT:
    fprintf(output,"%s=0",f->name);
    return 1;
  case NEXT:
    print_tree_wring_infix(f,output);
    return 1;
  case AND:
    l=print_tree_wring(f->left,output);
    if(l)
      fprintf(output,",");
    r=print_tree_wring(f->right,output);
    return l+r;
  }
  
}  



/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
    
/**Function********************************************************************

  Synopsis    [It converts the tree in NNF.]

******************************************************************************/
static tree* NNF_recur(tree* t,int flag){
  if(t==NULL)
    return NULL;
  else if(t->type==NOT){
    tree* tmp=NNF_recur(t->left,!flag);
    myfree(t);
    return tmp;
  } else {
    if(!flag)
      switch(t->type){
      case TRUE:
	t->type=FALSE;
	break;
      case FALSE:
	t->type=TRUE;
	break;
      case ATOM:
	t->type=NOT;
	break;
      case NEXT:
	t->type=NEXT;
	break;
      case AND:
	t->type=OR;
	break;
      case OR:
	t->type=AND;
	break;
      case UNTIL:
	t->type=RELEASE;
	break;
      case RELEASE:
	t->type=UNTIL;
	break;
      }
    t->left=NNF_recur(t->left,flag);
    t->right=NNF_recur(t->right,flag);
    return t;
  }
    
}

/**Function********************************************************************

  Synopsis    [It converts a tree into a formula.]

******************************************************************************/
static form* tree2form_recur(tree* t){
  form* f;
  switch(t->type){
  case TRUE:
  case FALSE:
  case ATOM:
  case NOT:
    f=new_form(t->type,t->prop,0,0);
    free(t);
    return f;
  case NEXT:
    f=new_form(t->type,t->prop,tree2form_recur(t->left),0);
    free(t);
    return f;
  case AND:
  case OR:
  case UNTIL:
  case RELEASE:
    f=new_form(t->type,t->prop,tree2form_recur(t->left),tree2form_recur(t->right));
    free(t);
    return f;
  }
}
